"""
Provides testing data sets.

SHAPES will always be:
inputs X: [n_batch, n_timesteps, n_states]
outputs y: [n_batch, n_timesteps, n_states]

"""

import numpy as np

# TODO: add Lorenz and other test cases (with potential limits to the number of states etc.)


def gen_sine(n=10, omega=np.pi):
    # generates a sequence of sines and cosines with a given frequency. sampling is 10 points per period
    t = np.linspace(start=0, stop=n / 50 * omega, num=n, endpoint=True)
    return np.sin(omega * t)


def gen_cos(n=10, omega=np.pi):
    # generates a sequence of sines and cosines with a given frequency. sampling is 10 points per period
    t = np.linspace(start=0, stop=n / 50 * omega, num=n, endpoint=True)
    return np.cos(omega * t)


def gen_sincos(n=10, omega=np.pi, a_sc=1, b_sc=0.25, P_sc=3):
    # generates a sequence of a_sc*sin(omega*t)^P_sc + b_sc*cos(omega*t)^P_sc
    # using gen_sine and gen_cos functions
    sine_wave = gen_sine(n, omega)
    cos_wave = gen_cos(n, omega)
    return a_sc * sine_wave**P_sc + b_sc * cos_wave**P_sc


def split_sequence(signal, n_batch, n_time_in, n_time_out):
    # expects [n_timesteps, n_states] sequence

    # convert into inputs (function at last n_time_in points) and outputs (function at next n_time_out points)
    x, y = [], []
    for i in range(n_batch):
        idx_in_end = i + n_time_in
        idx_out_end = idx_in_end + n_time_out
        x.append(signal[i:idx_in_end, :])  # last n_timestep_in points
        y.append(signal[idx_in_end:idx_out_end, :])  # next n_timestep_out points
    x, y = np.array(x), np.array(y)

    return x, y


def train_test_split(x, y):
    # train-test split 80% sample random points from the sequence and return inputs and outputs
    n = x.shape[0]

    ratio = 0.8

    split_idx = np.max([1, int(n * ratio)])

    shuffle_idx = np.random.choice(n, size=n, replace=False)
    train_idx, test_idx = shuffle_idx[: int(n * ratio)], shuffle_idx[int(n * ratio) :]

    # split data according to train/test split and return.
    return x[train_idx], x[test_idx], y[train_idx], y[test_idx]


def sine_pred(n_batch, n_time_in, n_time_out, n_states):
    # predict a sine signal. Single- and multi-step prediction supported

    # we will create different signal frequencies for the different states
    signal, omega = [], np.pi
    for _ in range(n_states):
        signal.append(gen_sine(n=n_batch + (n_time_in + n_time_out) + 1, omega=omega))
        omega += 0.314

    # 2D array of shape [n_timesteps, n_states]
    signal = np.array(signal).transpose()

    # split into inputs and outputs
    x, y = split_sequence(signal, n_batch, n_time_in, n_time_out)

    # train-test split
    X_train, X_test, y_train, y_test = train_test_split(x, y)

    return X_train, X_test, y_train, y_test


def sine_to_cosine(n_batch, n_time_in, n_time_out, n_states):
    # Generate sine input and cosine output signals
    total_time = n_batch + n_time_in + n_time_out
    x, y = [], []
    omega = np.pi
    for _ in range(n_states):
        x.append(gen_sine(n=total_time, omega=omega))
        y.append(gen_cos(n=total_time, omega=omega))
        omega += 0.314

    # Convert to 2D arrays of shape [n_timesteps, n_states]
    x = np.array(x).T
    y = np.array(y).T

    # Split into sequences
    x = split_sequence(x, n_batch, n_time_in, n_time_out)
    y = split_sequence(y, n_batch, n_time_in, n_time_out)

    # Unpack the tuples returned by split_sequence
    x_input, _ = x
    _, y_output = y

    # Train-test split
    X_train, X_test, y_train, y_test = train_test_split(x_input, y_output)

    return X_train, X_test, y_train, y_test


def sincos2(n_batch, n_time_in, n_time_out, n_states):
    # Generate sine input and SinCos-2 output signals
    total_time = n_batch + n_time_in + n_time_out
    x, y = [], []
    omega = 1  # As specified in the document

    for _ in range(n_states):
        x.append(gen_sine(n=total_time, omega=omega))
        y.append(gen_sincos(n=total_time, omega=omega))
        omega += 0.314  # Increment omega for each state, as in the original function

    # Convert to 2D arrays of shape [n_timesteps, n_states]
    x = np.array(x).T
    y = np.array(y).T

    # Split into sequences
    x = split_sequence(x, n_batch, n_time_in, n_time_out)
    y = split_sequence(y, n_batch, n_time_in, n_time_out)

    # Unpack the tuples returned by split_sequence
    x_input, _ = x
    _, y_output = y

    # Train-test split
    X_train, X_test, y_train, y_test = train_test_split(x_input, y_output)

    return X_train, X_test, y_train, y_test


"""
CASE 1: Vector to Vector
"""


def vector_to_vector(name, n_batch: int = 50, n_states=1):
    assert type(n_states) is int
    n_time_in = 1

    # make sure to have at least 1 testing sample
    n_batch = np.max([n_batch, 2])

    if name == "sine_prediction":
        # single-step predict a sine signal
        X_train, X_test, y_train, y_test = sine_pred(
            n_batch=n_batch, n_states=n_states, n_time_in=1, n_time_out=1
        )

    elif name == "sine_to_cosine":
        # Map a sequence of sines to a sequence of cosines
        X_train, X_test, y_train, y_test = sine_to_cosine(
            n_batch=n_batch, n_states=n_states, n_time_in=1, n_time_out=n_time_out
        )

    elif name == "sin_to_cos2":
        # Map a sequence of sines to a sequence of sinecosines
        X_train, X_test, y_train, y_test = sincos2(
            n_batch=n_batch,
            n_states=n_states,
            n_time_in=n_time_in,
            n_time_out=n_time_out,
        )

    print(f"shape of inputs (training): {X_train.shape}")
    print(f"shape of outputs (training): {y_test.shape}")
    return X_train, X_test, y_train, y_test


"""
CASE 2: Sequence to scalar
"""


def sequence_to_scalar(name, n_batch: int = 50, n_states=1, n_time_in=2):
    assert type(n_states) is int
    assert type(n_time_in) is int
    n_time_out = 1

    # make sure to have at least 1 testing sample
    n_batch = np.max([n_batch, 2])

    if name == "sine_prediction":
        # single-step predict a sine signal
        X_train, X_test, y_train, y_test = sine_pred(
            n_batch=n_batch,
            n_states=n_states,
            n_time_in=n_time_in,
            n_time_out=n_time_out,
        )

    elif name == "sine_to_cosine":
        # Map a sequence of sines to a sequence of cosines
        X_train, X_test, y_train, y_test = sine_to_cosine(
            n_batch=n_batch,
            n_states=n_states,
            n_time_in=n_time_in,
            n_time_out=n_time_out,
        )

    elif name == "sin_to_cos2":
        # Map a sequence of sines to a sequence of sinecosines
        X_train, X_test, y_train, y_test = sincos2(
            n_batch=n_batch,
            n_states=n_states,
            n_time_in=n_time_in,
            n_time_out=n_time_out,
        )

    print(f"shape of inputs (training): {X_train.shape}")
    print(f"shape of outputs (training): {y_test.shape}")
    return X_train, X_test, y_train, y_test


"""
CASE 3: Sequence to sequence
"""


def sequence_to_sequence(name, n_batch: int = 50, n_states: int = 2, n_time: int = 3):

    # make sure to have at least 1 testing sample
    n_batch = np.max([n_batch, 2])

    if name == "sine_pred":
        # multi-step predict a vector of sine signals
        X_train, X_test, y_train, y_test = sine_pred(
            n_batch=n_batch,
            n_states=n_states,
            n_time_in=n_time,
            n_time_out=n_time,
        )
    elif name == "sine_to_cosine":
        # Map a sequence of sines to a sequence of cosines
        X_train, X_test, y_train, y_test = sine_to_cosine(
            n_batch=n_batch,
            n_states=n_states,
            n_time_in=n_time,
            n_time_out=n_time,
        )

    elif name == "sin_to_cos2":
        # Map a sequence of sines to a sequence of sinecosines
        X_train, X_test, y_train, y_test = sincos2(
            n_batch=n_batch,
            n_states=n_states,
            n_time_in=n_time,
            n_time_out=n_time,
        )

    print(f"shape of inputs (training): {X_train.shape}")
    print(f"shape of outputs (training): {y_test.shape}")
    return X_train, X_test, y_train, y_test


def x_to_x(
    name,
    n_batch: int = 50,
    n_states_in: int = 2,
    n_states_out: int = 2,
    n_time_int: int = 1,
    n_time_out: int = 1,
):

    # make sure to have at least 1 testing sample
    n_batch = np.max([n_batch, 2])

    # full flexibility in creating input and output shapes
    n_states = np.max([n_states_in, n_states_out])

    if name == "sine_pred":
        # single-step predict a vector of sine signals
        X_train, X_test, y_train, y_test = sine_pred(
            n_batch=n_batch, n_states=n_states, n_time_in=1, n_time_out=1
        )

    elif name == "sine_to_cosine":
        # Map a sequence of sines to a sequence of cosines
        X_train, X_test, y_train, y_test = sine_to_cosine(
            n_batch=n_batch, n_states=n_states, n_time_in=1, n_time_out=1
        )

    elif name == "sincos2":
        # Map a sequence of sines to a sequence of cosines
        X_train, X_test, y_train, y_test = sincos2(
            n_batch=n_batch, n_states=n_states, n_time_in=1, n_time_out=1
        )

    # cut data if input and output vector length is not the same
    X_train, X_test = X_train[:, :, :n_states_in], X_test[:, :, :n_states_in]
    y_train, y_test = y_train[:, :, :n_states_out], y_test[:, :, :n_states_out]

    print(
        f"shape of inputs (training): {X_train.shape}, shape of outputs (training): {y_test.shape}"
    )

    return X_train, X_test, y_train, y_test


if __name__ == "__main__":

    # case 1

    # n_time = X_train.shape[1]
    # plt.figure()
    # for i in range(3):
    #     plt.plot(np.arange(start=1, stop=n_time), X_train[i,:,0], label='input')
    #
    # plt.legend()
    # plt.show()

    X_train, X_test, y_train, y_test = scalar_to_scalar(name="sincos2", n_batch=50)

    # print(X_train,X_test, y_test,y_train)

    X_train, X_test, y_train, y_test = vector_to_vector(
        name="sincos2", n_batch=1, n_states=3
    )

    # print(X_train,X_test, y_test,y_train)

    X_train, X_test, y_train, y_test = sequence_to_sequence(
        name="sincos2", n_batch=50, n_states=4, n_time=15
    )

    # print(X_train,X_test, y_test,y_train)

    X_train, X_test, y_train, y_test = x_to_x(
        name="sincos2",
        n_batch=100,
        n_states_in=4,
        n_states_out=3,
        n_time_int=10,
        n_time_out=2,
    )



"""
We will have an abstract Layer class, from which the following layers inherit:

- InputLayer
- ReservoirLayer
    - RandomReservoir
    - RecurrenceReservoir
    - EvolvedReservoir
- ReadoutLayer

"""

from abc import ABC, abstractmethod
import numpy as np
import networkx as nx

from pyreco.utils_networks import (
    gen_ER_graph,
    compute_density,
    get_num_nodes,
    compute_spec_rad,
    remove_nodes_from_graph,
)


# implements the abstract base class
class Layer(ABC):

    @abstractmethod
    def __init__(self):
        self.weights = None  # every layer will have some weights (trainable or not)
        self.name: str = "layer"
        pass

    @abstractmethod
    def update_layer_properties(self):
        pass


class InputLayer(Layer):
    # Shape of the read-in weights is: N x n_states, where N is the number of nodes in the reservoir, and n_states is
    # the state dimension of the input (irrespective if a time series or a vector was put in)
    # the actual read-in layer matrix will be created by mode.compile()!

    def __init__(self, input_shape):
        # input shape is (n_timesteps, n_states)
        super().__init__()
        self.shape = input_shape
        self.n_time = input_shape[0]
        self.n_states = input_shape[1]
        self.name = "input_layer"

        # some properties of the readin layer
        self.fraction_nonzero_entries: (
            float  # fraction of nonzero entries in the input layer
        )

    def remove_nodes(self, nodes: list):
        # removes a node from the input layer (i.e. if a reservoir node needs to be dropped)

        if not isinstance(nodes, list):
            raise TypeError("Nodes must be provided as a list of indices.")
        if np.max(nodes) > self.weights.shape[0]:
            raise ValueError(
                "Node index exceeds the number of nodes in the input layer."
            )
        if np.min(nodes) < 0:
            raise ValueError("Node index must be positive.")
        if not all(isinstance(x, int) for x in nodes):
            raise ValueError("All entries in the node list must be integers.")

        # remove nodes from [n_reservoir_nodes, n_states] matrix
        self.weights = np.delete(self.weights, nodes, axis=0)

        # update the properties of the input layer
        self.update_layer_properties()

    def update_layer_properties(self):

        # updates the properties of the input layer
        self.fraction_nonzero_entries = (
            np.count_nonzero(self.weights) / self.weights.size
        )


class ReadoutLayer(Layer):

    def __init__(self, output_shape, fraction_out=1.0):
        # expects output_shape = (n_timesteps, n_states)
        super().__init__()
        self.output_shape: tuple = output_shape
        self.n_time = output_shape[0]
        self.n_states = output_shape[1]

        self.fraction_out: float = (
            fraction_out  # fraction of connections to the reservoir
        )
        self.name = "readout_layer"

        self.readout_nodes = []  # list of nodes that are linked to output

    def remove_nodes(self, nodes: list):
        # remove nodes from the readout layer, when the user wants to delete specific reservoir nodes

        # # delete entries that equal the nodes to be removed
        # idx_del = np.where(np.isin(self.readout_nodes, nodes))[0]
        # self.readout_nodes = np.delete(self.readout_nodes, idx_del)
        # self.update_layer_properties()
        raise NotImplementedError("This method is not yet implemented.")

    def update_layer_properties(self):

        # updates the properties of the readout layer
        self.fraction_out = len(self.readout_nodes) / self.weights.shape[0]
        pass


class ReservoirLayer(Layer):  # subclass for the specific reservoir layers

    def __init__(
        self,
        nodes,
        density,
        activation,
        leakage_rate,
        fraction_input,
        init_res_sampling,
        seed: int = 42,
    ):
        super().__init__()
        self.nodes: int = nodes  # number of reservoir nodes
        self.density: float = density
        self.spec_rad = None
        self.activation = activation
        self.leakage_rate = leakage_rate
        self.name = "reservoir_layer"
        self.fraction_input = fraction_input
        self.weights = None
        self.input_receiving_nodes = None

        # initial reservoir state (will be set later)
        self.initial_res_states = None
        self.init_res_sampling = init_res_sampling

    def activation_fun(self, x: np.ndarray):
        if self.activation == "sigmoid":
            return 1 / (1 + np.exp(-x))
        elif self.activation == "tanh":
            return (np.exp(x) - np.exp(-x)) / (np.exp(x) + np.exp(-x))
        else:
            raise (ValueError(f"unknown activation function {self.activation}!"))

    def set_weights(self, network: np.ndarray):
        # set reservoir network from outside.
        # Updates all related parameters

        self.weights = network

        # update reservoir properties
        self.update_layer_properties()

    def set_initial_state(self, r_init: np.ndarray):
        # assigns an initial state to each of the reservoir nodes

        if r_init.shape[0] != self.nodes:
            raise (
                ValueError(
                    "initial reservoir state does not match the number of nodes in the reservoir!"
                )
            )
        self.initial_res_states = r_init

    def remove_nodes(self, nodes: list):
        """
        Remove specified nodes from the reservoir network.
        Parameters:
        nodes (list): A list of indices representing the nodes to be removed.
        1. Removes the specified nodes from the adjacency matrix.
        2. Updates the reservoir properties including the number of nodes, density, and spectral radius.
        """
        # remove nodes from the reservoir network

        if not isinstance(nodes, list):
            raise TypeError("Nodes must be provided as a list of indices.")

        if np.max(nodes) > self.nodes:
            raise ValueError("Node index exceeds the number of nodes in the reservoir.")

        if np.min(nodes) < 0:
            raise ValueError("Node index must be positive.")

        # 1. remove nodes from the adjacency matrix
        self.weights = remove_nodes_from_graph(graph=self.weights, nodes=nodes)

        # 2. update reservoir properties
        self.update_layer_properties()

        # 3. update the initial reservoir state
        self.initial_res_states = np.delete(self.initial_res_states, nodes, axis=0)

        # # 4. update the info about input-receiving nodes
        # self.input_receiving_nodes = np.delete(self.input_receiving_nodes, nodes)

    def update_layer_properties(self):
        # Updates the reservoir properties including the number of nodes, density, and spectral radius.
        self.nodes = get_num_nodes(self.weights)
        self.density = compute_density(self.weights)
        self.spec_rad = compute_spec_rad(self.weights)


class RandomReservoirLayer(ReservoirLayer):
    def __init__(
        self,
        nodes,
        density: float = 0.1,
        activation: str = "tanh",
        leakage_rate: float = 0.5,
        fraction_input: float = 0.8,
        spec_rad: float = 0.9,
        init_res_sampling="random_normal",
        seed=None,
    ):

        # Call the parent class's __init__ method
        super().__init__(
            nodes=nodes,
            density=density,
            activation=activation,
            leakage_rate=leakage_rate,
            fraction_input=fraction_input,
            init_res_sampling=init_res_sampling,
            seed=seed,
        )

        # initialize subclass-specific attributes
        self.seed = seed
        self.spec_rad = spec_rad

        # generate a random ER graph using networkx
        self.weights = gen_ER_graph(
            nodes=nodes,
            density=density,
            spec_rad=self.spec_rad,
            directed=True,
            seed=seed,
        )

    def update_layer_properties(self):
        # Updates the reservoir properties including the number of nodes, density, and spectral radius.
        self.nodes = get_num_nodes(self.weights)
        self.density = compute_density(self.weights)
        self.spec_rad = compute_spec_rad(self.weights)


# class ReccurrenceLayer(ReservoirLayer):
#     # To Do: accept a random seed
#     def __init__(self, nodes, density, activation: str = 'tanh', leakage_rate: float = 0.2):
#         # Call the parent class's __init__ method
#         super().__init__(nodes, density, activation, leakage_rate)
#
#         # Initialize subclass-specific attributes
#         # https://pyts.readthedocs.io/en/stable/generated/pyts.image.RecurrencePlot.html#pyts.image.RecurrencePlot
#         # https://tocsy.pik-potsdam.de/pyunicorn.php
#



"""
Minimal testing example for predicting a sine function (discrete time setting)

x(t) -> x(t+dt)

"""

import numpy as np
from matplotlib import pyplot as plt

"""
generate some training and testing data from the sine function
"""

omega = np.pi
t = np.linspace(start=0, stop=5 * omega, num=51, endpoint=True)
signal = np.sin(omega * t)

# convert into inputs (function at current time point) and outputs (function at next time point)
x = signal[:-1]
y = signal[1:]

# train-test split 80%
ratio = 0.8
n = len(x)
shuffle_idx = np.random.choice(n, size=n, replace=False)
train_idx, test_idx = shuffle_idx[: int(n * ratio)], shuffle_idx[int(n * ratio) :]

# split data and create data shape [n_batch, n_timesteps, n_features]
X_train = np.expand_dims(x[train_idx], axis=(1, -1))
X_test = np.expand_dims(x[test_idx], axis=(1, -1))
y_train = np.expand_dims(y[train_idx], axis=(1, -1))
y_test = np.expand_dims(y[test_idx], axis=(1, -1))
print(f"shape of training data: {X_train.shape}, shape of test data: {y_test.shape}")

"""
Modeling using feed-forward neural nets in TensorFlow (sequential API)
"""

import tensorflow as tf
from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import Dense, Input

model_ann = Sequential()  # Instantiate Sequential model
model_ann.add(Input(shape=(1,)))  # Add input layer
model_ann.add(Dense(units=100, activation="sigmoid"))  # Add hidden Dense layer
model_ann.add(Dense(units=1, activation="linear"))  # Add output layer

model_ann.compile(optimizer="adam", loss="mean_squared_error")  # Compile the model
hist = model_ann.fit(
    X_train, y_train, epochs=500, batch_size=64
)  # Train the model for 1000 epochs
loss_ann = model_ann.evaluate(X_test, y_test)  # Evaluate the model
print(f"\n\n Dense Neural Network: test set loss: {loss_ann}")
y_pred_ann = model_ann.predict(X_test)  # Make predictions for new data


"""
Modeling using reservoir computing built on the pyreco library (ours)
"""
from pyreco.models import RC
from pyreco.layers import InputLayer, RandomReservoirLayer, ReadoutLayer

model_rc = RC()
model_rc.add(InputLayer(input_shape=(1, 1)))
model_rc.add(
    RandomReservoirLayer(
        nodes=100,
        density=0.1,
        activation="sigmoid",
        leakage_rate=0.2,
        fraction_input=0.5,
    )
)
model_rc.add(ReadoutLayer(output_shape=(1, 1), fraction_out=0.6))

# Compile the model
model_rc.compile(optimizer="ridge", metrics=["mean_squared_error"])

model_rc.fit(
    X_train, y_train
)  # , epochs=500)              # Train the model for 1000 epochs
y_pred_rc = model_rc.predict(X_test)  # Make predictions for new data

loss_rc = model_rc.evaluate(X_test, y_test)  # Evaluate the model
print(f"\n\n Reservoir Computer: test set loss: {loss_rc[0]}")



"""
Study the effect of reservoir network structure, i.e. the randomly generated reservoir adjacency matrix.

We will simply fit a large number of RCs, each with a randomly generated reservoir on a benchmark case and observe
the variation in the scores.

By Merten Stender, TU Berlin
merten.stender@tu-berlin.de
September 23, 2024
"""

import numpy as np
from matplotlib import pyplot as plt


from pyreco.models import ReservoirComputer
from pyreco.plotting import r2_scatter
from pyreco.utils_data import sequence_to_sequence

# some testing data: translate a sine signal to a cosine signal (i.e. sequence translation)
x_train, x_test, y_train, y_test = sequence_to_sequence(
    name="sine_to_cosine", n_batch=10, n_states=1, n_time=100
)
input_shape = (x_train.shape[1], x_train.shape[2])
output_shape = (y_train.shape[1], y_train.shape[2])


def run_trials(
    num_trials,
    x_train,
    y_train,
    x_test,
    y_test,
    num_nodes: int,
    activation: str = "tanh",
):
    # fits num_trials RCs to the same data,
    # returns the list of scores, as well as best and worst model

    test_scores = []
    best_score, worst_score = 10**6, 0
    best_model, worst_model = None, None
    for i in range(num_trials):
        print(f"network {i} / {num_trials}: {num_nodes} nodes, {activation} activation")

        # generate new model
        _model = ReservoirComputer(num_nodes=num_nodes, activation=activation)

        # fit to data
        _model.fit(X=x_train, y=y_train)

        # evaluate score
        _score = _model.evaluate(x_test, y_test, metrics="mae")[0]
        test_scores.append(_score)

        # store worst and best model
        if _score > worst_score:
            print(f"current score {_score:.4f} worse than before ({worst_score:.4f})")
            worst_model = _model
            worst_score = _score

        if _score < best_score:
            print(f"current score {_score:.4f} better than before ({best_score:.4f})")
            best_model = _model
            best_score = _score

        del _model

    test_scores = np.array(test_scores)  # turn into np array for easier handling

    return np.array(test_scores), best_model, worst_model


if __name__ == "__main__":

    # run the analysis for different numbers of reservoir nodes. Hypothesis: the larger the number of reservoir
    # nodes, the less will be the effect of network structure (redundancy of information)

    num_trials = 100  # for each RC specification in terms of hyperparameters
    num_nodes = [10, 50, 100, 250, 500]  # number of nodes to try out
    activation = "tanh"  # common activation function
    model_scores, best_models, worst_models = [], [], []

    for _nodes in num_nodes:
        # get scores for all trials for the given number of nodes
        _scores, _best_model, _worst_model = run_trials(
            num_trials=num_trials,
            x_train=x_train,
            y_train=y_train,
            x_test=x_test,
            y_test=y_test,
            num_nodes=_nodes,
            activation=activation,
        )
        model_scores.append(_scores)
        best_models.append(_best_model)
        worst_models.append(_worst_model)

    # plot the variation in the training performance based on reservoir initialization
    plt.figure()
    for idx, _nodes in enumerate(num_nodes):
        plt.hist(model_scores[idx], bins=20, label=rf"$N=${_nodes}", alpha=0.3)
    plt.xlabel("test score (MSE, sine to cosine)")
    plt.xlim([0, None])
    plt.ylabel("counts")
    plt.title(f"variation of reservoir network, {num_trials} different networks")
    plt.legend()
    plt.savefig("effects_network_structure_histogram.png")
    # plt.show()

    # Create the figure and axis
    fig, ax = plt.subplots()
    ax.boxplot(model_scores, showmeans=True, meanline=True)  # Create the boxplot
    ax.set_xticklabels(num_nodes)  # Set the x-axis labels
    ax.set_xlabel(r"reservoir nodes $N$")
    ax.set_ylabel(r"test error ($L_1$)")
    ax.set_title(f"Performance variation by random networks ({num_trials} trials)")
    plt.savefig("effects_network_structure_boxplot.png")
    # plt.show()

    # plot predictions of best and worst model for first test sample
    plt.figure()
    plt.plot(y_test[0, :, 0], label="ground truth", color="black")
    for idx, _nodes in enumerate(num_nodes):
        y_pred_best = best_models[idx].predict(x_test)
        y_pred_worst = worst_models[idx].predict(x_test)
        plt.plot(
            y_pred_best[0, :, 0], label=f"best model (N={_nodes})", linestyle="dashed"
        )
        # plt.plot(y_pred_worst[0, :, 0], label=f'worst model (N={_nodes})', linestyle='dashed')
    plt.xlabel("time")
    plt.ylabel("target / prediction")
    plt.title(rf"variation induced by network structure")
    plt.legend()
    plt.savefig("effects_network_structure_predictions.png")
    # plt.show()



"""
Some testing use case for the CustomModel API of pyreco
"""

from matplotlib import pyplot as plt
from pyreco.custom_models import RC as RC
from pyreco.layers import InputLayer, ReadoutLayer
from pyreco.layers import RandomReservoirLayer
from pyreco.plotting import r2_scatter
from pyreco.utils_data import sequence_to_scalar
from pyreco.optimizers import RidgeSK

"""
Classic RC built on random networks. 

Use case: predict a sine signal into the future (n_time) samples from the past are mapped to (n_time) samples in the 
future
"""

# some testing data: predict a sine signal from a sequence of 20 function values
X_train, X_test, y_train, y_test = sequence_to_scalar(
    name="sine_prediction",
    n_states=1,
    n_batch=200,
    n_time_in=20,
)

# set the dimensions (need to be [n_batch, n_time, n_states])
input_shape = (X_train.shape[1], X_train.shape[2])
output_shape = (y_train.shape[1], y_train.shape[2])

model_rc = RC()
model_rc.add(InputLayer(input_shape=input_shape))
model_rc.add(
    RandomReservoirLayer(
        nodes=200, density=0.1, activation="tanh", leakage_rate=0.1, fraction_input=0.5
    ),
)
model_rc.add(ReadoutLayer(output_shape, fraction_out=0.99))

# Compile the model
optim = RidgeSK(alpha=0.5)
model_rc.compile(
    optimizer=optim,
    metrics=["mean_squared_error"],
)

# Train the model
model_rc.fit(X_train, y_train)

# Make predictions for new data
y_pred = model_rc.predict(X_test)
print(f"shape of predictions on test set: {y_pred.shape}")

# Evaluate the model
loss_rc = model_rc.evaluate(X_test, y_test, metrics=["mae"])
print(f"Test model loss: {loss_rc}")

# plot predictions vs. ground truth
r2_scatter(y_true=y_test, y_pred=y_pred)


# plot predictions
plt.figure()
plt.plot(X_test[0, :, 0], color="gray", alpha=0.3)
plt.plot(X_test[0, :, 0], marker=".", color="black", label="input sequence")
plt.plot(
    X_test.shape[1], y_test[0, 0, 0], marker="o", color="blue", label="true target"
)
plt.plot(X_test.shape[1], y_pred[0, 0, 0], marker="x", color="red", label="predicted")
plt.xlabel("time")
plt.ylabel("amplitude")
plt.legend()
plt.title("sequence to scalar prediction")
plt.show()


"""
k-fold cross-validation of the model: estimate bias and variance of the model
"""
from pyreco.cross_validation import cross_val

val, mean, std_dev = cross_val(model_rc, X_train, y_train, n_splits=5, metric=["mse"])
print(f"Cross-Validation MSE: {val}")
print(f"Mean MSE: {mean:.3f}")
print(f"Standard Deviation of MSE: {std_dev:.3f}")
